class ReleaseNotes:

    __PendingPromotionCaption = "Pending promotion"
    __conf = {}
    __ticketsByVersion = {}
    __promotedVersionsInfo = {}
    __ticketReplacements = {}
    __ticketProvider = {}

    def __init__(self, conf, ticketProvider, repo, promotedVersionsInfo):
        self.__ticketsByVersion = {}
        self.__conf = conf
        self.__repo = repo
        self.__promotedVersionsInfo = promotedVersionsInfo
        self.__ticketProvider = ticketProvider;

        self.__repo.checkout()
        self.__repo.retrieveHistory()
        self.__repo.retrieveVersionsByGitHash( list(self.__promotedVersionsInfo.keys()) )
        self.__computeTicketsByVersion()
        
    def __computeTicketsByVersion(self):
        currentVersion = "latest"        
        self.__ticketsByVersion[currentVersion] = []
        for hash in self.__repo.gitCommitsList:
            if hash in self.__repo.versionsByGitHash:
                currentVersion = self.__repo.versionsByGitHash[hash]
                self.__ticketsByVersion[currentVersion] = []
            tickets = self.__ticketProvider.extractTicketsFromMessage(self.__repo.gitCommitMessagesByHash[hash])
            self.__ticketsByVersion[currentVersion] = self.__ticketsByVersion[currentVersion] + tickets
     
    def __printVersionBlock(self, version, tickets, writer):
        date = "N/A"
        deps = {}
        if version != self.__PendingPromotionCaption:
            if version in self.__promotedVersionsInfo:
                date = self.__promotedVersionsInfo[version]["date"]
                if "directDependencies" in self.__promotedVersionsInfo[version]:
                    deps = self.__promotedVersionsInfo[version]["directDependencies"]

        if len(tickets) == 0:
            return ""

        return writer.printVersionBlock(deps, version, date, tickets)

    def generateReleaseNotesByPromotedVersions(self, writer):
        
        ticketsSoFar = []
        hashesInVersion = self.__repo.gitHistoryByVersion

        hashAlreadySeen = []
        sortedVersions = [] + list(hashesInVersion.keys())
        sortedVersions.sort(key=lambda s: list(map(int, s.split('.'))))
        
        content = []
        for version in sortedVersions:
            if version in self.__promotedVersionsInfo or len(self.__promotedVersionsInfo.keys()) == 0:
                print("Generating info for version " + version)
                
            for hash in hashesInVersion[version]:
                if hash in hashAlreadySeen:
                    continue
                hashAlreadySeen = hashAlreadySeen + [hash]
                if hash not in self.__repo.gitCommitMessagesByHash:
                    self.__repo.processCommit(hash)
                ticketsSoFar += self.__ticketProvider.extractTicketsFromMessage(self.__repo.gitCommitMessagesByHash[hash])

            if version in self.__promotedVersionsInfo or len(self.__promotedVersionsInfo.keys()) == 0:
                content = [self.__printVersionBlock(version, ticketsSoFar, writer)] + content
                ticketsSoFar = []                

        if len(ticketsSoFar) > 0:
            content = [self.__printVersionBlock(self.__PendingPromotionCaption, ticketsSoFar, writer)] + content

        return "\n".join(content).strip(' \t\n\r')



import os

class dll:

    def extractVersions(self, fileContent, fileName):
        if fileContent is None:
            return None
        dependencyVersion = self.__calcversioninfo(fileContent)
        res = []
        res.append("{0}: {1}".format(os.path.basename(fileName), dependencyVersion))
        return res


    def __normalizer(self, s):
        for j in range(len(s)):
            if len(s[j]) > 3:
                k = s[j][2:]
            else:
                k = '0' + s[j][2:]
            s[j] = k
        return s

    def __calcversioninfo(self, fn):
        verstrings = []
        sigstrings = self.__findsignatures(fn)
        if sigstrings[0] == '':
            return None
        for i in sigstrings:
            FV = self.__normalizer(i.split(',')[8:16])
            FOS = self.__normalizer(i.split(',')[32:36])
            hexver = FV[3]+FV[2]+FV[1]+FV[0]+':'+FV[7]+FV[6]+FV[5]+FV[4]
            if hexver not in verstrings:
               verstrings.append(hexver)
        myver = max(verstrings)
        return self.__parsver(myver)

    def __createparsestruct(self, b):
        s= ''
        for i in range(len(b)):
            byte = b[i]
            if type(byte) is str:
                byte = ord(byte)
            s += hex(byte)+','
        return s[:-1]

    def __findsignatures(self, sz):
        res = []
        indx=sz.find(b'\xbd\x04\xef\xfe')
        cnt = sz.count(b'\xbd\x04\xef\xfe')
        while cnt > 1:
            s = self.__createparsestruct(sz[indx:indx+52])
            sz = sz[indx+1:]
            cnt = sz.count(b'\xbd\x04\xef\xfe')
            indx=sz.find(b'\xbd\x04\xef\xfe')
            res.append(s)
        res.append(self.__createparsestruct(sz[indx:indx+52]))
        return res

    def __parsver(self, v):
        a,b,c,d = v[:4], v[4:8], v[9:13], v[13:]
        return str(int(a,16)) + '.'+ str(int(b,16)) +'.' + str(int(c,16)) + '.' + str(int(d,16))



import JustReleaseNotes
from JustReleaseNotes import factory
from JustReleaseNotes.issuers import AggregateIssuer


def create(conf):
    if isinstance(conf, list):
        return JustReleaseNotes.issuers.AggregateIssuer.AggregateIssuer(conf)
    else:
        return JustReleaseNotes.factory.createWithConfig("JustReleaseNotes.issuers", conf["Provider"], conf)



import sys
import requests
import json
from JustReleaseNotes.issuers import BaseIssues

class GitHubIssues(BaseIssues.BaseIssues):
    __conf = None
    __cache = {}
    
    def __init__(self, conf):
        self.__conf = conf
        self.__iconMappings = dict(
            issue="http://www.ic.gc.ca/app/opic-cipo/trdmrks/srch/imageLoader?appNum=1366861&extension=",
            pull_request="https://addons.cdn.mozilla.net/user-media/addon_icons/603/603460-64.png?modified=1428920625")
        self.ticketRegex = '#([0-9]+)'
        if "TicketRegex" in conf:
            self.ticketRegex = conf["TicketRegex"]

        headers = { 'Authorization': self.__conf["Authorization"] }
        response = requests.get( self.__conf["Url"] + "?filter=all&state=all", headers = headers, verify=False )
        tickets = json.loads(response.text)

        for ticketData in tickets:
            ticketNumber = str(ticketData["number"])
            self.__cache[ticketNumber] = ticketData

    def __log(self, message):
        print ("GitHub Issues: " + message)
        sys.stdout.flush()
        
    def __readJsonInfo(self, ticket):

        if ticket in self.__cache.keys():
            self.__log("Cached ticket info for " + ticket)
            return self.__cache[ticket]
        else:
            uri = self.__conf["Url"] + "/" + ticket
            self.__log("Retrieving ticket info for {0}: {1}".format(ticket, uri))

            headers = { 'Authorization': self.__conf["Authorization"] }
            r = requests.get( uri, headers = headers, verify=False )

            data = json.loads(r.text)
            if "message" in data and data["message"] == "Not Found":
                self.__log("Error retrieving GitHub Issue info: " + data["message"])
            self.__cache[ticket] = data
            return data

    def getTicketInfo(self, ticket):
        data = self.__readJsonInfo(ticket)

        if "title" in data:
            title = data["title"]
            embedded_links = {}
            for t in self.extractTicketsFromMessage(title):
                embedded_links["#" + t] = "{0}/{1}".format(self.__conf["HtmlUrl"], t)
        else:
            title = "Untitled"

        if "pull_request" in data:
            issueType = "pull_request"
        else:
            issueType = "issue"

        return { "issue_type_icon" : self.__iconMappings[issueType],
            "html_url" : data["html_url"],
            "ticket" : "#{0}".format(ticket),
            "title" : title,
            "embedded_link" : embedded_links,
            "reporter": data["user"]["login"]}




import re
import os

def replace(config):
    if type(config) is dict:
        return __replaceDict(config)
    elif type(config) is list:
        return __replaceList(config)
    elif type(config) is int:
        return config
    elif type(config) is float:
        return config
    else:
        return __replaceStr(str(config))

def __replaceDict(config):
    result = {}
    for key in config:
        result[key] = replace(config[key])
    return result


def __replaceList(config):
    return [replace(x) for x in config]


def __replaceStr(str):
    return re.sub(r'ENV\[(\w+)\]', __findEnvKey, str, flags=re.IGNORECASE)


def __findEnvKey(match):
    return os.environ[match.group(1)]


import sys
import JustReleaseNotes
from JustReleaseNotes.issuers import BaseIssues

class AggregateIssuer(BaseIssues.BaseIssues):
    __conf = None
    __issuers = {}
    
    def __init__(self, issuersConfArray):
        self.__issuers = []
        for issuerConf in issuersConfArray:
            issuer = JustReleaseNotes.issuers.factory.create(issuerConf)
            self.__issuers.append(issuer)

    def extractTicketsFromMessage(self, message):
        ret = []
        for issuer in self.__issuers:
            ret.extend(issuer.extractTicketsFromMessage(message))
        return ret

    def getTicketInfo(self, ticket):
        for issuer in self.__issuers:
            if issuer.extractTicketsFromMessage(ticket) != ["NULL"]:
                return issuer.getTicketInfo(ticket)
        return None



import re
import sys
from JustReleaseNotes.writers import BaseWriter

class MarkdownWriter(BaseWriter.BaseWriter):

    def __init__(self, ticketProvider):
        BaseWriter.BaseWriter.__init__(self, ticketProvider)

    def getExtension(self):
        return ".md"

    def getImageBlock(self, icon):
        return "![Icon]({0})".format(icon)

    def getVersionHeader(self, version):
        return "## {0} ##".format(version)

    def parseVersionHeader(self, line):
        if (line.startswith("## ") and line.endswith(" ##")):
            return line[3:len(line)-3]
        else:
            return False

    def printVersionBlock(self, deps, version, date, tickets):
        baseoutput = BaseWriter.BaseWriter.printVersionBlock(self, deps, version, date, tickets)
        if baseoutput is not None:
            return baseoutput

        version = self.convertVersion(version)
        data = [self.getVersionHeader(version)]
        if date != 'N/A':
            data.append(date)

        data.append("")
        uniqTickets = sorted(set(tickets), reverse=True)
        appendStabilityImprovements = False;

        for ticket in uniqTickets:
            if ticket == "NULL":
                appendStabilityImprovements = True
            else:
                ticketInfo = self.ticketProvider.getTicketInfo(ticket)
                if ticketInfo != None:
                    title = ticketInfo["title"]
                    if "embedded_link" in ticketInfo:
                        for ticket, link in list(ticketInfo["embedded_link"].items()):
                            title = re.sub(ticket, "[{1}]({0})".format(link, ticket), title)
                    iconPart = ""
                    if "issue_type_icon" in ticketInfo:
                        iconPart = self.getImageBlock(ticketInfo["issue_type_icon"])
                    data.append("* {0} [{1}]({3}) {2}, *reported by* **{4}**".format(iconPart, ticketInfo["ticket"],
                                                              title.encode('ascii','ignore').decode("ascii") ,
                                                              ticketInfo["html_url"],
                                                              ticketInfo["reporter"]))

        if appendStabilityImprovements:
            data.append("* Stability improvements")

        return '\n'.join(data) + '\n'


import sys
import requests
import json
import io

class GitHubReleases:
    __defaultVersionRegex = "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"
    __conf = []
    __promotedVersions = {}
        
    def __init__(self, conf):
        self.__promotedVersions = {}
        self.__conf = conf

    def __log(self, message):
        log = "GitHub Releases: " + message
        print(log)
        sys.stdout.flush()

    def retrievePromotedVersions(self):
        self.__log("Retrieving promoted from {0} at {1} ...".format(
            self.__conf["Provider"],
            self.__conf["Url"]))

        headers = { 'Authorization': self.__conf["Authorization"] }
        r = requests.get( self.__conf["Url"], headers = headers, verify=False )

        response = json.loads(r.text)
        for release in response:
            version = release["name"]
            self.__promotedVersions[version] = {
                "date" : release["published_at"],
                }
        
        if len(self.__promotedVersions) == 0:
            raise ValueError("Failed to retrieve promoted version. Please check if configuration is correct")
        
        self.__log("Found {0} promoted versions".format(len(self.__promotedVersions)))
        return self.__promotedVersions



from JustReleaseNotes.writers import MarkdownWriter

class StashMarkdownWriter(MarkdownWriter.MarkdownWriter):

    def __init__(self, ticketProvider):
        MarkdownWriter.MarkdownWriter.__init__(self, ticketProvider)

    def getImageBlock(self, icon):
        return ""



import re
from JustReleaseNotes.writers import BaseWriter

class HtmlWriter(BaseWriter.BaseWriter):

    def __init__(self, ticketProvider):
        BaseWriter.BaseWriter.__init__(self, ticketProvider)

    def getExtension(self):
        return ".html"

    def getVersionHeader(self, version):
        return "<div style=\"width:100%; border: 0px\"><a name=\"{0}\" class=\"version\"></a>".format(version)

    def parseVersionHeader(self, line):
        if (line.startswith("<div style=\"width:100%; border: 0px\"><a name=\"") and line.endswith("\" class=\"version\"></a>")):
            return line[46:len(line)-22]
        else:
            return False


    def printVersionBlock(self, deps, version, date, tickets):
        baseoutput = BaseWriter.BaseWriter.printVersionBlock(self, deps, version, date, tickets)
        if baseoutput is not None:
            return baseoutput

        version = self.convertVersion(version)

        data = [self.getVersionHeader(version), "<h2>" + version]

        if date != 'N/A':
            data.append("<sup><small style=\"font-size:10px\"><i> " + date + "</i></small></sup>")
        data.append("</h2>")

        if len(deps.keys()) > 0:
            data.append('<div style="background: #eee; "><i>Components: ')
            data.append('; '.join(deps.values()))
            data.append('</i></div>')

        data.append("<ul>")
        uniqTickets = sorted(set(tickets), reverse=True)

        appendStabilityImprovements = False

        for ticket in uniqTickets:
            if ticket == "NULL":
                appendStabilityImprovements = True
            else:
                ticketInfo = self.ticketProvider.getTicketInfo(ticket)
                if ticketInfo != None:
                    title = ticketInfo["title"]
                    if "embedded_link" in ticketInfo:
                        for ticket, link in list(ticketInfo["embedded_link"].items()):
                            title = re.sub(ticket, "<a href=\"{0}\">{1}</a>".format(link, ticket), title)
                    imgFormat = '<img src="{1}" alt="{0}" width="16" height="16" style="padding-right: 5px"></img>'
                    imgHtml = ""
                    if "state_icon" in ticketInfo:
                        imgHtml += imgFormat.format("State", ticketInfo["state_icon"])
                    if "issue_type_icon" in ticketInfo:
                        imgHtml += imgFormat.format("Issue Type", ticketInfo["issue_type_icon"])
                    if "priority_icon" in ticketInfo:
                        imgHtml += imgFormat.format("Priority", ticketInfo["priority_icon"])
                    data.append('<li style="font-size:14px">{0}<a href="{3}">{1}</a> {2}, <i>reported by</i> <b>{4}</b></li>'.format(
                        imgHtml, ticketInfo["ticket"], title , ticketInfo["html_url"], ticketInfo["reporter"]))

        if appendStabilityImprovements:
            data.append("<li style=\"font-size:14px\">Stability improvements</li>")

        data += ["</ul>", "</div>", ""]

        return '\n'.join(data)


from JustReleaseNotes.writers import MarkdownWriter

class GithubMarkdownWriter(MarkdownWriter.MarkdownWriter):

    def __init__(self, ticketProvider):
        MarkdownWriter.MarkdownWriter.__init__(self, ticketProvider)

    def getImageBlock(self, icon):
        return "<img src=\"{0}\" width=16 height=16></img>".format(icon)



import sys
import re
import json
import tempfile
import os
import JustReleaseNotes.artifacters.versioners
from JustReleaseNotes.artifacters.versioners import factory
import requests


class Artifactory:
    __artifactoryUrl = None
    __defaultVersionRegex = "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"
    __conf = []
    __promotedVersions = {}
        
    def __init__(self, conf):
        self.__promotedVersions = {}
        self.__conf = conf
        self.__artifactoryUrl = self.__conf["StorageUrl"]

    def __log(self, message):
        print ("Artifactory: " + message)
        sys.stdout.flush()
        
    def __retrievePromotionDate(self, version):
        self.__log("Retrieving info for version {0}".format(version))
        uri = "{0}/{1}/{2}/{3}".format(
            self.__artifactoryUrl,
            self.__conf["Repository"],
            self.__conf["ArtifactUri"],
            version)
        r = requests.get( uri )
        response = json.loads(r.text)
        
        if "created" in response:
            return response["created"][:10]
        else:
            self.__log("No 'created' for version {0}".format(version))
            
        return "n/a"
        
    def __downloadFile(self, uri):
        r = requests.get( uri )
        if r.status_code == 404:
            return None
            
        response = json.loads(r.text)
        if not "downloadUri" in response:
            self.__log("Failed to get 'downloadUri' from response.")
            return None
            
        self.__log("Download URI: {0}".format(response["downloadUri"]))
        
        #        
        r = requests.get(response["downloadUri"], stream=True)
        if r.status_code != 200:
            return None

        b = bytes()
        for chunk in r.iter_content(1024):
            b += chunk
        return b

    def retrieveDependeciesVersions(self, version):
        if "DirectDependencies" not in self.__conf:
            return {}
    
        if len(self.__conf["DirectDependencies"].keys()) == 0:
            return {}
        
        result = {}
        for packageName in list(self.__conf["DirectDependencies"].keys()):
            
            dep = self.__conf["DirectDependencies"][packageName]

            versionsExtractor = JustReleaseNotes.artifacters.versioners.factory.create(dep["type"])
            if packageName == "ANY":
                if dep["type"] == "ivy":
                    uri = "{0}/{1}/{2}/ivy-{2}.xml".format(
                        self.__conf["Repository"],
                        self.__conf["ArtifactUri"],
                        version).replace("//","/")
                    fullUrl = "{0}/{1}".format(self.__artifactoryUrl,uri)
                    result[packageName] = '; '.join(versionsExtractor.extractVersions(self.__downloadFile(fullUrl), fullUrl))
                else:
                   raise ValueError( "Unsupported dependency type '{0}' for ANY".format(dep["type"]))
            else:
                # Extract from DLL
                if dep["type"] == "dll":
                    uri = "{0}/{1}/{2}/{3}".format(
                        self.__conf["Repository"],
                        self.__conf["ArtifactUri"],
                        version,
                        dep["name"]).replace("//","/")
                    fullUrl = "{0}/{1}".format(self.__artifactoryUrl,uri)
                    dependencyVersions = versionsExtractor.extractVersions(self.__downloadFile(fullUrl), fullUrl)
                    if dependencyVersions is None:
                        continue
                    result[packageName] = '; '.join(dependencyVersions)
                else:
                    raise ValueError( "Unsupported dependency type '{0}'".format(dep[packageName]["type"]))
        
        return result
          
    def __extractVersion(self, version):
        versionRegex = self.__defaultVersionRegex
        if "RegexToExtractVersion" in self.__conf:
            versionRegex = self.__conf["RegexToExtractVersion"]
        regex = re.compile(versionRegex)

        extractedVersion = regex.findall(version)      
        if len(extractedVersion) == 0:
            self.__log("Failed to find version in '{0}' using '{1}'".format(version, versionRegex))
            return None
        if len(extractedVersion) > 1:
            self.__log("Multiple versions are matching the regex for: '{0}'".format(version))
            return None
            
        return extractedVersion[0].replace("_",".")
    

    def retrievePromotedVersions(self):
        self.__log("Retrieving promoted ({0}) versions {1} ...".format(
            self.__conf["Repository"],
            self.__conf["ArtifactUri"]))

        uri = "{0}/{1}".format(self.__artifactoryUrl,self.__conf["Repository"]);
        artifactPath=self.__conf["ArtifactUri"];
        if (artifactPath[0] == "/"):
            uri += artifactPath
        else:
            uri += "/" + artifactPath
            
        r = requests.get( uri )
        response = json.loads(r.text)
        
        for child in response["children"]:
            version = child["uri"][1:]

            extractedVersion = self.__extractVersion(version)
            if extractedVersion is None:
                continue
            
            self.__promotedVersions[extractedVersion] = {
                "date" : self.__retrievePromotionDate(version),
                "directDependencies" : self.retrieveDependeciesVersions(version)
                }
        
        if len(self.__promotedVersions) == 0:
            raise ValueError("Failed to retrieve promoted version. Please check artifactory configuration is correct")
        
        self.__log("Found {0} promoted versions".format(len(self.__promotedVersions)))
        return self.__promotedVersions


import re

class BaseIssues():

    ticketRegex = None

    def extractTicketsFromMessage(self, message):
        message = message.replace("\n", " ").replace("\r", " ").replace("\t", " ");
        p = re.compile(self.ticketRegex)
        results = p.findall(message)
        if len(results) > 0:
            return results
        else:
            return ["NULL"]


import importlib
import sys

def createWithConfig(module, name, conf):
    try:
        module = importlib.import_module(module + ".{0}".format(name))
        clazz = getattr(module, name)
        return clazz(conf)
    except:
        e = sys.exc_info()[1]
        raise Exception("Unable to create module provider {0} from module {1}: {2}".format(conf["Provider"], module, e))

def createWithName(module, name):
   try:
        module = importlib.import_module(module + ".{0}".format(name))
        clazz = getattr(module, name)
        return clazz()
   except:
        e = sys.exc_info()[1]
        raise Exception("Unable to create module provider {0} from module {1}: {2}".format(name, module, e))


import xml.etree.ElementTree as ET

class ivy:

    def extractVersions(self, fileContent, fileName):
        root = ET.fromstring(fileContent)
        deps = root.findall('./dependencies/dependency')
        res = []
        for dep in deps:
            v = dep.attrib['name'] + ": " + dep.attrib['rev']
            if 'revConstraint' in dep.attrib:
                v = v + " (" + dep.attrib['revConstraint'] + ")"
            res.append(v)
        return res


import os, sys, re
from git import Repo
        
class GitRepo:
    __repo = ""
    __repoX = ""
    __packageName = ""
    __remote = "origin"
    __branch = "master"
    
    gitCommitsList = []
    gitCommitMessagesByHash = {}
    gitDatesByHash = {}
    versionsByGitHash = {}
    commitParents = {}
    gitHistoryByVersion = {}
    
    def __init__(self, conf):
        self.gitCommitsList = []
        self.gitCommitMessagesByHash = {}
        self.gitDatesByHash = {}
        self.versionsByGitHash = {}
        self.gitHistoryByVersion = {}
        self.__repo = conf["RepositoryUrl"]
        self.__directory = conf["Directory"]
        self.__repoX = ""
        self.__versionTagRegex = "^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$"

        if "VersionTagRegex" in conf:
            self.__versionTagRegex = conf["VersionTagRegex"]

        if "Remote" in conf:
            self.__remote = conf["Remote"]

        if "Branch" in conf:
            b = conf["Branch"]
            pos = b.find("/")
            if (pos >= 0):
                self.__remote = b[:pos]
                self.__branch = b[pos+1:]
            else:
                self.__branch = b


        
    def __log(self, message):
        print ("Git: " + message)
        sys.stdout.flush()
        
    def checkout(self):
        path = self.__directory
        if not os.path.isdir(path):
            self.__log("Creating folder at: " + path)
            os.makedirs(path)

        self.__log("Cloning " + self.__repo + " at " + path)
        try:
            self.__repoX = Repo.clone_from(self.__repo, path)
            release_notes_head = self.__repoX.create_head(self.__branch, self.__remote + "/" + self.__branch)
            self.__repoX.head.reference = release_notes_head
            self.__repoX.head.reset(index=True, working_tree=True)
        except:
            self.__repoX = Repo(path)
            self.__repoX.head.reference = self.__repoX.heads[self.__branch]

    def setParents(self, commit):
        if len(commit.parents) == 0:
            return
            
        if commit.hexsha in self.commitParents:
            # Already traversed
            return

        self.commitParents[commit.hexsha] = []
        for p in commit.parents:
            self.commitParents[commit.hexsha] = self.commitParents[commit.hexsha] + [p.hexsha]
            self.setParents(p)
            
    def __getParentsListForVersion(self, hash, version, resultsSoFar):
        if hash in resultsSoFar:
            return []
        if hash not in self.commitParents:
            return []
        results = [hash]
        for p in self.commitParents[hash]:
            if p not in self.versionsByGitHash:
                results = results + self.__getParentsListForVersion(p, version, resultsSoFar + results)
        return results

    def processCommit(self, commitHash):
        self.__processCommit(self.__repoX.commit(commitHash))

    def __processCommit(self, commit):
        self.gitCommitMessagesByHash[commit.hexsha] = commit.summary + commit.message
        self.gitCommitsList.append(commit.hexsha)
        self.gitDatesByHash[commit.hexsha] = commit.authored_date
        self.setParents(commit)

    def retrieveHistory(self):
        self.__log("Retrieving Git history...")        
        for i in self.__repoX.iter_commits(self.__branch, max_count=1024):
            self.__processCommit(i)
        
    def __optimizeHistoryByVersion(self):        
        sortedVersionsInAscendingOrder = [] + list(self.gitHistoryByVersion.keys())
        sortedVersionsInAscendingOrder.sort(key=lambda s: list(map(int, s.split('.'))), reverse=False)
        
        # remove commits part of newer versions if they exists in older one        
        hashUsedForVersion = {}
        for version in sortedVersionsInAscendingOrder:
            for hash in self.gitHistoryByVersion[version]:
                if hash in hashUsedForVersion:
                    x = [] + self.gitHistoryByVersion[version]
                    x.remove(hash)
                    self.gitHistoryByVersion[version] = [] + x
                else:
                    hashUsedForVersion[hash] = version
        
    def retrieveVersionsByGitHash(self, promotedVersionsList):
        self.__log("Retrieving versions (git tags)...")

        if len(promotedVersionsList) == 0:
            print("Make sure you first retrieve the promoted version: Promoted versions are empty, thus every tag "
                  "matching will be considered as released")
        
        tags = self.__repoX.tags
        for t in tags:
            tag = "" + str(t)
            hexsha = str(t.commit)
            if tag.startswith("non-published"):
                continue

            version = tag.split("/")[-1]
            p = re.compile(self.__versionTagRegex)
            m = p.match(version)
            if not m:
                continue

            g = m.groups()
            if len(g)>1:
                continue;
            elif len(g) == 1:
                version = g[0];

            if hexsha in self.versionsByGitHash:
                v1 = version
                v2 = self.versionsByGitHash[hexsha]
                self.__log("Warning: Multiple versions for the same commit={0} v1={1} v2={2}".format(hexsha,v1,v2))
                self.versionsByGitHash[hexsha] = v1
                if v1 not in promotedVersionsList and v2 in promotedVersionsList:
                    self.versionsByGitHash[hexsha] = v2
                self.__log("Warning: Choosing {0} (promoted={1})".format(
                    self.versionsByGitHash[hexsha], 
                    self.versionsByGitHash[hexsha] in promotedVersionsList))
            else:
                self.versionsByGitHash[hexsha] = version     

            self.gitHistoryByVersion[version] = self.__getParentsListForVersion(hexsha,version, [])

        self.__addHeadIfNotPresent()
        self.__optimizeHistoryByVersion()

    def __addHeadIfNotPresent(self):
        hexsha = str(self.__repoX.heads[self.__branch].commit)
        if hexsha not in self.versionsByGitHash:
            version = str(sys.maxsize)
            self.versionsByGitHash[hexsha] = version
            self.gitHistoryByVersion[version] = self.__getParentsListForVersion(hexsha, version, [])


import re
import sys

class BaseWriter:

    def __init__(self, ticketProvider):
        self.ticketProvider = ticketProvider
        self.versionsAlreadyPresent = {}
        self.__upcomingDevelopments = "Upcoming developments"

    def convertVersion(self, version):
        if (version == str(sys.maxsize)):
            version = self.__upcomingDevelopments
        return version

    def deconvertVersion(self, version):
        if (version == self.__upcomingDevelopments):
            version = str(sys.maxsize)
        return version

    def parseVersionHeader(self, line):
        return False

    def setInitialContent(self, content):
        currentVersion = None
        for line in content.split("\n"):
            v = self.deconvertVersion(self.parseVersionHeader(line))
            if v is not False and v != str(sys.maxsize):
                currentVersion = v
                self.versionsAlreadyPresent[v] = []
            if currentVersion is not None:
                self.versionsAlreadyPresent[currentVersion].append(line)
        return self.versionsAlreadyPresent

    def printVersionBlock(self, deps, version, date, tickets):
        if version in self.versionsAlreadyPresent.keys():
            return '\n'.join(self.versionsAlreadyPresent[version]);
        return None



import os
from setuptools import setup

description = "Release notes generator package"
cur_dir = os.path.dirname(__file__)
try:
    long_description = open(os.path.join(cur_dir, 'README.rst')).read()
except:
    long_description = description

setup(name='JustReleaseNotes',
      version='0.1.8',
      description='Release notes generator package',
      long_description=long_description,
      url='https://github.com/Cimpress-MCP/JustReleaseNotes',
      author='Ivan Stanishev, Rafal Nowosielski',
      author_email='ivan@stanishev.info, rafal@nowosielski.link',
      license='Apache License 2.0',
      packages=['JustReleaseNotes', 'JustReleaseNotes.artifacters', 'JustReleaseNotes.issuers',
                'JustReleaseNotes.artifacters.versioners','JustReleaseNotes.writers', 'JustReleaseNotes.sourcers',
                'JustReleaseNotes.utils'],
      install_requires=[
          'requests',
          'gitpython'
      ],
      zip_safe=False,
      entry_points="""
      [console_scripts]
      just_release = JustReleaseNotes.command_line:main
      """
)



import json
import os
import os.path
import argparse

import requests

import JustReleaseNotes.artifacters
import JustReleaseNotes.writers
import JustReleaseNotes.issuers
import JustReleaseNotes.sourcers
from JustReleaseNotes.sourcers import factory
from JustReleaseNotes.issuers import factory
from JustReleaseNotes.writers import factory
from JustReleaseNotes.artifacters import factory
from JustReleaseNotes.releaseNotes import *
from JustReleaseNotes.utils import EnvReplacer

def main():
    parser = argparse.ArgumentParser(prog="just_release", description='Instruments release process.')
    parser.add_argument('command', metavar='cmd', action="store", choices=['notes'])
    parser.add_argument('--config', '-c', '--c', metavar='config.json', nargs='?', help='Specifies the configuration to use', default="config.json")
    args = parser.parse_args()

    if args.command == "notes":
        generate_release_notes(args.config)

def generateForOneWriter(generator, ticketProvider, writerType, directory, fileName):
    print("\nGenerating using {0}".format(writerType))

    writer = JustReleaseNotes.writers.factory.create(writerType, ticketProvider)

    if fileName is None:
        fileName = "index{0}".format(writer.getExtension())

    content = ""
    p = os.path.join(directory, fileName)
    if os.path.isfile(p):
        f = open(p, "r")
        content = f.read()
        f.close()

    writer.setInitialContent(content)
    releaseNotes = generator.generateReleaseNotesByPromotedVersions(writer)

    print("\nStoring release notes at {0}".format(os.path.join(directory, fileName)))
    f = open(os.path.join(directory, fileName), "wb")
    f.write(releaseNotes.encode('utf-8'))
    f.close()

def generate_release_notes(configFile):
        requests.packages.urllib3.disable_warnings()
        file = open(configFile, 'r')
        fileContents = file.read()
        releaseNotesConfig = EnvReplacer.replace(json.loads(fileContents))
        currentDir = os.getcwd()

        if not os.path.isabs(releaseNotesConfig["pathToSave"]):
            releaseNotesConfig["pathToSave"] = os.path.join(currentDir, releaseNotesConfig["pathToSave"])

        for packageName, conf in releaseNotesConfig["packages"].items():
            if "Releases" in conf:
                releasesConf = conf["Releases"]
                promotedVersionsInfo = JustReleaseNotes.artifacters.factory.create(releasesConf).retrievePromotedVersions()
            else:
                print ("No artifacter configured: every version tag will be considered a valid release")
                promotedVersionsInfo = {}

            issuesConf = conf["Issues"]
            ticketProvider = JustReleaseNotes.issuers.factory.create(issuesConf)

            directory = os.path.join(releaseNotesConfig["pathToSave"],packageName)

            if "Directory" in conf["Source"]:
                if not os.path.isabs(conf["Source"]["Directory"]):
                    conf["Source"]["Directory"] = os.path.join(currentDir, conf["Source"]["Directory"])
            else:
                conf["Source"]["Directory"] = directory

            repo = JustReleaseNotes.sourcers.factory.create(conf["Source"])

            if not os.path.exists(conf["Source"]["Directory"]):
                os.makedirs(conf["Source"]["Directory"])

            generator = ReleaseNotes(conf, ticketProvider, repo, promotedVersionsInfo)
            writerConfigs = conf["ReleaseNotesWriter"]
            if isinstance(writerConfigs, str):
                generateForOneWriter(generator, ticketProvider, writerConfigs, directory, None)
            elif hasattr(writerConfigs, '__iter__'):
                for writerConf in writerConfigs:
                    path = writerConf["PathToSave"]
                    generateForOneWriter(generator, ticketProvider, writerConf["Provider"], os.path.dirname(path), os.path.basename(path))

if __name__ == '__main__':
    main()



import sys, re
import requests
import json
from JustReleaseNotes.issuers import BaseIssues

class JiraIssues(BaseIssues.BaseIssues):
    __restSearchUrl = None
    __jiraAuthorization = None
    __cache = {}

    def __init__(self, conf):
        self.__restSearchUrl = conf["Url"]
        self.__jiraAuthorization = conf["Authorization"]
        self.__conf = conf
        self.ticketRegex = '([A-Z]{2,5}-[0-9]+)'
        if "TicketRegex" in conf:
            self.ticketRegex = conf["TicketRegex"]

    def __log(self, message):
        print ("Jira: " + message)
        sys.stdout.flush()

    def __readJsonInfo(self, ticket):

        if ticket in self.__cache:
            self.__log("Cached ticket info for " + ticket)
            return self.__cache[ticket]

        uri = "{0}/{1}".format(self.__restSearchUrl,ticket)
        headers = { 'Authorization': self.__jiraAuthorization }
        self.__log("Retrieving ticket info for " + ticket)
        r = requests.get( uri, headers = headers, verify=False )
        data = json.loads(r.text)
        if "errorMessages" in data:
            self.__log("Error retrieving Jira info: " + ",".join(data["errorMessages"]))
        self.__cache[ticket] = data
        return data

    def getTicketInfo(self, ticket):
        data = self.__readJsonInfo(ticket)

        embedded_links = {}
        title = "Untitled"
        ret = { "html_url" : "{0}/{1}".format(self.__conf["HtmlUrl"],ticket),
                "ticket" : ticket,
                "title" : title }

        if "fields" in data.keys():
            title = data["fields"]["summary"]
            ret["title"] = title
            for ticket in self.extractTicketsFromMessage(title):
                embedded_links[ticket] = "{0}/{1}".format(self.__conf["HtmlUrl"],ticket)
            ret["state_icon"] = self.__fieldIcon(data["fields"]["status"])
            ret["issue_type_icon"] = self.__fieldIcon(data["fields"]["issuetype"])
            ret["priority_icon"] = self.__fieldIcon(data["fields"]["priority"])
            ret["embedded_link"] = embedded_links
            ret["reporter"] = data["fields"]["reporter"]["displayName"]
        else:
            return None

        return ret

    def __fieldIcon(self, f):
        if "WebImagesPath" in self.__conf:
            parts = f["iconUrl"].split("/")
            return '{0}/{1}'.format(self.__conf["WebImagesPath"], parts[len(parts)-1], f["name"])
        else:
            return f["iconUrl"]
